from typing import Any, Optional, Type, TypeVar

TItem = TypeVar("TItem", bound=object)


class ProviderCache:
    __instance: Optional["ProviderCache"] = None
    __objects: dict[Type, Any] = {}

    @classmethod
    def get_instance(cls) -> "ProviderCache":
        if cls.__instance is None:
            cls.__instance = cls()
        return cls.__instance

    def __iadd__(self, other: TItem) -> "ProviderCache":
        self.__objects[other.__class__] = other
        return self

    def __getitem__(self, ttype: Type[TItem]) -> TItem:
        return self.__objects[ttype]

    def __setitem__(self, ttype: Type[TItem], object: TItem) -> None:
        if ttype not in object.__class__.__mro__:
            raise ValueError(f"Object of type {object.__class__} cannot be set under type {ttype}")
        self.__objects[ttype] = object

    @classmethod
    def flush(cls):
        cls.__instance = None
        cls.__objects = {}



from typing import Any, Optional, Type, TypeVar

TItem = TypeVar("TItem", bound=object)


class AutoWiredCache:
    __instance: Optional["AutoWiredCache"] = None

    @classmethod
    def get_instance(cls, evaluate_lazy: bool = False) -> "AutoWiredCache":
        if cls.__instance is None:
            cls.__instance = cls(evaluate_lazy)
        return cls.__instance

    def __init__(self, evaluate_lazy: bool = False) -> None:
        """
        :param evaluate_lazy: If True, the peer dependencies will be evaluated only when they are requested.
        """
        self.evaluate_lazy = evaluate_lazy
        self.__objects: dict[Type, Any] = {}
        self.__blueprints: set[Type] = set()
        self.__evaluated_deps: set[Type] = set()

    def __iadd__(self, other: TItem | Type[TItem]) -> "AutoWiredCache":
        if type(other) == type:
            self.__set_dependency(other)
        else:
            self.__set_instantiated_dependency(other.__class__, other)
        return self

    def __setitem__(self, ttype: Type[TItem], object: TItem) -> None:
        if ttype not in object.__class__.__mro__:
            raise ValueError(f"Object of type {object.__class__} cannot be set under type {ttype}")
        self.__set_instantiated_dependency(ttype, object)

    def __getitem__(self, ttype: Type[TItem]) -> TItem:
        if ttype not in self.__objects:
            return self.__instantiate(ttype)
        return self.__objects[ttype]

    def __set_dependency(self, ttype: Type) -> None:
        if self.evaluate_lazy:
            self.__blueprints.add(ttype)
            return

        self.__evaluate_deps(ttype)
        self.__blueprints.add(ttype)

    def __set_instantiated_dependency(self, ttype: Type, object: Any) -> None:
        self.__objects[ttype] = object
        self.__blueprints.add(ttype)
        self.__evaluated_deps.add(ttype)

    def __evaluate_deps(self, ttype: Type[TItem]) -> None:
        annotations = self.__get_annotes(ttype)

        all_deps = list({*annotations.values()}.difference(self.__evaluated_deps))
        for peer_dep in all_deps:
            if peer_dep not in self.__blueprints:
                raise ValueError(f"Peer dependency {peer_dep} not found for {ttype}")

            self.__evaluated_deps.add(peer_dep)
            new_peer_deps = {*self.__get_annotes(peer_dep).values()}
            all_deps.extend(new_peer_deps.difference(self.__evaluated_deps))

    def __instantiate(self, ttype: Type[TItem]) -> TItem:
        if ttype not in self.__blueprints:
            raise ValueError(f"Object of type {ttype} not found or could not be instantiated.")

        self.__evaluate_deps(ttype)

        annotations = self.__get_annotes(ttype)
        initialization_kwargs = {
            field: self.__objects.get(field_type) or self.__instantiate(field_type)
            for field, field_type in annotations.items()
            if field != "return"
        }
        self.__objects[ttype] = ttype(**initialization_kwargs)
        return self.__objects[ttype]

    @staticmethod
    def __get_annotes(ttype: Type[TItem]) -> dict[str, Type]:
        return dict(filter(lambda x: x[0] != "return", ttype.__init__.__annotations__.items()))

    @classmethod
    def flush(cls):
        cls.__instance = None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dependency_resolution']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'dependency-resolution',
    'version': '0.2.1',
    'description': 'A simple dependency resolution library using container concepts',
    'long_description': '# Dependency Resolution\n\n## How to use\n\n### ProviderCache Usage\n\nWhen autowiring is not required use ProviderCache for a simple implementation.\n\nBecause autowiring is not involved there are algorithms to determine the order of initialization. And therefore ProviderCache is faster.\n\n```python\nfrom dependency_resolution import ProviderCache\n\nclass Image:\n  def __init__(self, file: str):\n    self.file = file\n\nclass ImageProcessor:\n  def __init__(self, image: Image):\n    self.image = image\n\n  def sharpen(self):\n    pass\n\ndef init_image():\n  cache = ProviderCache.get_instance()\n  cache += Image("image.png")\n  # Can also be done like this\n  # cache[Image] = Image("image.png")\n\ndef init_processor():\n  cache = ProviderCache.get_instance()\n  cache += ImageProcessor(cache[Image])\n\ndef process_image():\n  cache = ProviderCache.get_instance()\n  processor = cache[ImageProcessor]\n  processor.sharpen()\n\nif __name__ == "__main__":\n  init_image()\n  init_processor()\n  process_image()\n```\n\n### AutoWiredCache Usage\n\nWhen autowiring is required use AutoWiredCache.\n\nAutoWiredCaches accept just the types of the dependencies and will automatically resolve them (given required deps are already present in the Cache or can be created using other deps).\n\n```python\nfrom dependency_resolution import AutoWiredCache\n\nclass ImageDriver:\n  def __init__(self, proc: ImageProcessor, image: Image):\n    self.proc = proc\n    self.image = image\n\n  def edit_and_save(self):\n    pass\n\ndef init_image():\n  cache = AutoWiredCache.get_instance()\n  cache += Image("image.png")\n  # Can also be done like this\n  # cache[Image] = Image("image.png")\n\ndef init_processor_and_driver():\n  cache = AutoWiredCache.get_instance()\n  cache += ImageProcessor\n  cache += ImageDriver\n\ndef process_image():\n  cache = AutoWiredCache.get_instance()\n  driver = cache[ImageDriver]\n  driver.edit_and_save()\n\nif __name__ == "__main__":\n  init_image()\n  init_processor()\n  process_image()\n```\n\n`AutoWiredContainer` also supports lazy evaluation of dependencies. When a new type is added to the cache, it is not initialized immediately and as a result any missing deps will be ignored at that moment.\n\nTherefore, `init_processor_and_driver` function from the above example can be as follows\n\n```python\ndef init_processor_and_driver():\n  cache = AutoWiredCache.get_instance()\n  cache += ImageDriver\n  cache += ImageProcessor\n```\n\nWhen the `ImageDriver` object is obtained from the cache (`cache[ImageDriver]`), the `ImageProcessor` object will be created first following which the `ImageDriver` object will be created.\n\n## Future Plans\n\n- [x] Add autowire support\n- [ ] Add Providers to pass specific arguments to the constructor\n- [ ] Add support for deleting or removing provided instances in both `ProviderCache` and `AutoWiredCache`\n',
    'author': 'Saroopashree K',
    'author_email': 'saroopa25@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8.0,<4.0.0',
}


setup(**setup_kwargs)
